import logging

import requests

default_session = requests.Session()
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)

log = logging.getLogger()


def fetch_json(url, r):
    respone_data = {}
    if r is None:
        return respone_data
    if r.status_code != 200:
        log.error('http fail: %s %s %s', url, r.status_code, r.text);
        return respone_data
    try:
        respone_data = r.json()
    except ValueError:
        log.error('http json_decode error: %s %s', url, r.text);
    return respone_data


def get(url, session=None, **kwargs):
    session = default_session if session is None else session
    call = lambda: session.get(url, **kwargs)
    return safe_reqeust(url, call)


def post(url, session=None, **kwargs):
    session = default_session if session is None else session
    call = lambda: session.post(url, **kwargs)
    return safe_reqeust(url, call)


def safe_reqeust(url, call):
    try:
        r = call()
        return r
    except requests.exceptions.RequestException as e:
        return None



# -*- coding: utf-8 -*-
import unittest
import tempfile
import sys
from os import path

import shutil

from echobox.tool import template
from echobox.tool import file
from echobox.tool import strings

from echobox.tool.testutils import RandomDir


class TestTempalte(unittest.TestCase, RandomDir):

    def setUp(self):
        RandomDir.setUp(self)
        self.template_dir = path.join(self.tests_root_dir, 'test-data')

    def tearDown(self):
        RandomDir.tearDown(self)

    def test_write(self):
        import six
        target_file = path.join(self.temp_dir, self.random_str(10))
        strs = ['b', u'b', '好', u'好']
        for s in strs:
            file.file_put_contents(target_file, s)
            ret = file.file_get_contents(target_file)
            if six.PY2:
                self.assertEqual(type(ret), unicode)
            else:
                self.assertEqual(type(ret), str)


if __name__ == '__main__':
    unittest.main()



import unittest

from echobox.tool import system
from echobox.tool import file
from echobox.tool.testutils import RandomDir
from echobox.tool.datatypes import NameIdDict


class TestSystem(unittest.TestCase, RandomDir):

    def setUp(self):
        RandomDir.setUp(self)

    def tearDown(self):
        RandomDir.tearDown(self)

    def test_all(self):
        cmd = 'git symbolic-ref HEAD 2>/dev/null | cut -d"/" -f 3'
        self._check(cmd)

        cmd = 'some command unknown'
        self._check(cmd)

    def _check(self, cmd):
        ret = system.run_cmd(cmd)

        file_name = self.temp_dir + '/shell-run.out'
        with open(file_name, 'w') as stdout_file:
            returncode = system.shell_run(cmd, stdout=stdout_file, exit_on_error=False, stderr=stdout_file)
        output = file.file_get_contents(file_name).rstrip('\n')
        self.assertEqual(ret[0], returncode)
        self.assertEqual(ret[1], output)


if __name__ == '__main__':
    unittest.main()



from echobox.tool.utils import Singleton


class ConfigData():
    __metaclass__ = Singleton

    def __init__(self):
        self._env = None
        self._app_name = None
        self._app_config = {}

    def get_env(self):
        return self._env

    def get_app_name(self):
        return self._app_name

    def init(self, app_name, env=None, app_config={}):
        self._app_name = app_name
        if env is not None:
            self._env = env
        self._app_config = app_config

    def get_app_config(self):
        return self._app_config


appconfig = ConfigData()



import os
import shutil
from os import path
import six
import hashlib
from echobox.tool import system

_STR_MAP = {
    ' ': '.',
    '/': '-',
    '--': '-',
}


def listdir(dir, relative_path=False):
    for root, directories, filenames in os.walk(dir):
        for filename in filenames:
            filepath = path.join(root, filename)
            if relative_path:
                filepath = filepath.replace(dir, '')
            yield filepath


def ensure_dir(dir, mode=0o777):
    if not os.path.exists(dir):
        oldmask = os.umask(000)
        os.makedirs(dir, mode)
        os.umask(oldmask)


def ensure_dir_for_file(file, mode=0o777):
    dir = os.path.dirname(file)
    ensure_dir(dir, mode)


def compute_lock_filepath(sys_argv):
    filepath = ' '.join(sys_argv)
    for old, new in _STR_MAP.items():
        filepath = filepath.replace(old, new)
    return filepath + '.lock'


def obtain_lock(filepath):
    ensure_dir_for_file(filepath)
    pidfile = open(filepath, 'a+')
    try:
        if system.get_os_name() == 'windows':
            import portalocker
            portalocker.lock(pidfile.fileno(), portalocker.LOCK_EX | portalocker.LOCK_NB)
        else:
            import fcntl
            fcntl.flock(pidfile.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        pidfile = None
        return None
    pidfile.seek(0)
    pidfile.truncate()
    pidfile.write(str(os.getpid()))
    pidfile.flush()
    pidfile.seek(0)
    return pidfile


def rmtree(folder_path):
    for file_object in os.listdir(folder_path):
        file_object_path = os.path.join(folder_path, file_object)
        if os.path.isfile(file_object_path) or os.path.islink(file_object_path):
            os.unlink(file_object_path)
        else:
            shutil.rmtree(file_object_path)


def copytree(src, dst, symlinks=False, ignore=None):
    for file_object in os.listdir(src):
        src_file_object_path = os.path.join(src, file_object)
        dst_file_object_path = os.path.join(dst, file_object)
        if os.path.isfile(src_file_object_path):
            shutil.copy(src_file_object_path, dst_file_object_path)
        elif os.path.islink(src_file_object_path):
            linkto = os.readlink(src_file_object_path)
            os.symlink(linkto, dst_file_object_path)
        else:
            shutil.copytree(src_file_object_path, dst_file_object_path, symlinks, ignore)


def free_space(path):
    if not os.path.isdir(path):
        return 0
    s = os.statvfs(path)
    free_space = (s.f_bavail * s.f_frsize)
    return free_space


def file_get_contents(file, mode='r'):
    with open(file, mode) as f:
        data = f.read()
        data = ensure_encode(data)
        return data


def ensure_encode(data):
    # for python 2, byte string should be convert to unicode
    if six.PY2 and type(data) == str:
        data = data.decode('utf-8')
    # for python 3, all string, type `str` is unicode
    return data


def ensure_decode(data):
    # for python 2, unicode should be encode to byte string, so that we can write them into file
    if six.PY2 and type(data) == unicode:
        data = data.encode('utf-8')
    return data


def file_put_contents(target_file, data, mode='w'):
    ensure_dir_for_file(target_file)
    data = ensure_decode(data)
    with open(target_file, mode) as f:
        f.write(data)


def md5_file(file):
    if not os.path.isfile(file):
        return ''
    hash_md5 = hashlib.md5()
    with open(file, 'rb') as f:
        for chunk in iter(lambda: f.read(4096), b''):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



import os
import socket
import sys
from os import path

from echobox.app.appconfig import appconfig
from echobox.tool import array
from echobox.tool import file
from echobox.tool import functocli
from echobox.tool import logger
from echobox.tool import system
from echobox.tool import template
from echobox.tool import utils
from echobox.tool import spec
from echobox.tool import systemd

import __main__ as main


class DevOpsAppConfig:

    def __init__(self, app_name, provider):
        appconfig.init(app_name, provider.get_env(), provider.get_app_config())


class DevOpsAppConfigProvider:

    def __init__(self, app_name):
        self.app_name = app_name
        self.app_storage_dir = '/opt/data/' + app_name
        self.app_persistent_storage_dir = self.app_storage_dir + '/persistent'
        self.app_runtime_storage_dir = self.app_storage_dir + '/runtime'
        self.app_logs_dir = self.app_runtime_storage_dir + '/logs'

    def get_root_dir(self):
        if hasattr(main, '__file__'):
            root_dir = path.dirname(path.realpath(sys.argv[0]))
            return root_dir
        else:
            root_dir = os.getcwd()
            return root_dir

    def get_roles_dir(self):
        return self.get_root_dir() + '/roles'

    def get_app_root_dir(self):
        app_root_dir = self.get_roles_dir() + '/' + self.app_name
        return app_root_dir

    def get_env(self):
        env = os.environ['PUPPY_ENV'] if 'PUPPY_ENV' in os.environ else 'dev'
        return env

    def get_app_config(self):
        return {}

    def get_box_local_config(self):
        fn = self.get_root_dir() + '/.box-local-config.yml'
        return utils.load_yaml(fn, {})

    def ensure_dir_and_write_permission(self):
        file.ensure_dir(self.app_storage_dir)
        file.ensure_dir(self.app_persistent_storage_dir)
        file.ensure_dir(self.app_runtime_storage_dir)
        file.ensure_dir(self.app_logs_dir)
        file.ensure_dir(self.app_logs_dir + '/syslog')


class DevOpsApp(object):

    def _want_logger_ex(self, **kwargs):
        logger.make_logger_for_app(self, **kwargs)
        logger_name = kwargs.get('log_config', {}).get('name', self.app_name)
        return logger.getLogger(logger_name)

    def __init__(self, app_name, log_level='info', exit_on_error=True, **kwargs):
        self.app_name = app_name

        self._init_app_config()

        self.log_level = log_level
        self.exit_on_error = exit_on_error
        self.logger = self._want_logger_ex(**kwargs)

        self.file_lock = None
        self._hostvar = None

    def _init_app_config(self):

        provider = self.create_app_config_provider()
        provider.ensure_dir_and_write_permission()

        hostname = socket.gethostname()
        self.hostname_fqdn = hostname
        self.hostname_short = hostname.split('.', 1)[0]

        self.env = provider.get_env()
        self.root_dir = provider.get_root_dir()
        self.roles_dir = provider.get_roles_dir()

        app_root_dir = provider.get_app_root_dir()
        self.app_root_dir = app_root_dir
        self.app_config_dir = app_root_dir + '/config'
        self.app_templates_dir = app_root_dir + '/templates'
        self.app_scripts_dir = app_root_dir + '/scripts'

        self.app_storage_dir = provider.app_storage_dir
        self.app_persistent_storage_dir = provider.app_persistent_storage_dir
        self.app_runtime_storage_dir = provider.app_runtime_storage_dir
        self.app_logs_dir = provider.app_logs_dir


    def create_app_config_provider(self):
        return DevOpsAppConfigProvider(self.app_name)

    def run_cmd_ret(self, cmd, log=True):
        if log:
            self.logger.info('run_cmd_ret: %s', cmd)
        return system.run_cmd_ret(cmd)

    def run_cmd(self, cmd, log=True):
        if log:
            self.logger.info('run_cmd: %s', cmd)
        return system.run_cmd(cmd)

    def shell_run(self, cmd, keep_pipeline=True, exit_on_error=True, dry_run=False):
        self.logger.info('shell_run: %s', cmd)
        if dry_run:
            return 0
        exit_on_error = exit_on_error and self.exit_on_error
        return system.shell_run(cmd, keep_pipeline=keep_pipeline, exit_on_error=exit_on_error)

    def remove_container(self, name, force=False, dry_run=False):
        cmd = 'docker rm %s' % (name)
        if force:
            cmd = 'docker rm -f %s' % (name)
        self.shell_run(cmd, exit_on_error=False, dry_run=dry_run)

    def container_is_running(self, name):
        cmd = 'docker ps | grep %s' % (name)
        return self.run_cmd(cmd, log=False)[0] == 0

    def stop_container(self, name, timeout=300, dry_run=False):
        cmd = 'docker stop --time %d %s' % (timeout, name)
        self.shell_run(cmd, exit_on_error=False, dry_run=dry_run)

    def _check_lock(self):
        filepath = self.app_runtime_storage_dir + '/locks/' + file.compute_lock_filepath(sys.argv)
        file_lock = file.obtain_lock(filepath)
        if file_lock is None:
            pid = 0
            with open(filepath, 'r') as f:
                pid = f.read()
            self.logger.warning('lock file exists, pid: %s => %s', pid, filepath)
            sys.exit(1)
        else:
            self.file_lock = file_lock

    def template_to(self, template_filename, dst, payload, app_name=None):
        template_payload = {'payload': payload}
        src = self.app_templates_dir_for(app_name) + '/' + template_filename
        template.render_to_file(src, template_payload, dst)
        self.logger.info('template_to: %s => %s', src, dst)

    def read_app_config(self, config_filebasename, app_name=None):
        if '.yml' not in config_filebasename:
            config_filebasename = config_filebasename + '.yml'
        data = utils.load_yaml('%s/%s' % (self.app_config_dir_for(app_name), config_filebasename), {})
        return data

    def app_templates_dir_for(self, app_name=None):
        if app_name is None:
            return self.app_templates_dir
        return '%s/%s/templates' % (self.roles_dir, app_name)

    def app_config_dir_for(self, app_name=None):
        if app_name is None:
            return self.app_config_dir
        return '%s/%s/config' % (self.roles_dir, app_name)

    def get_group_names(self):
        return self.get_host_var().get('group_names', [])

    def get_configed_ops_scripts_dir(self):
        return self.get_host_var().get('ops_scripts_dir', '')

    def _read_hosts_var(self):
        fn = self.app_config_dir + '/hosts-var.yml'
        hosts_var = utils.load_yaml(fn, {})
        hostname = self.hostname_short
        return spec.LegacySpec(hosts_var, hostname).config_data

    def get_host_var(self):
        if self._hostvar is None:
            fn = '/etc/puppy-ansible-hostvar.json'
            ansible_kv_config = utils.load_json(fn, {})

            fn = '/etc/puppy-cloud-init-raw-data.json'
            cloud_init_raw_data = utils.load_json(fn, {})
            '''
            #### Always respect global_params / app_params from ansible
            1. Create an ECS by cloud-init, global_params from ansible, app_params from cloud-init config.
               We can not read app_params from ansible when we are using cloud-init to deloy,
               so we truncate `/etc/puppy-ansible-hostvar.json` when lauch an ECS.
               The global_params and app_params will be the value we desire.
            2. If we deploy by ansible, ansible_kv_config will overwrite cloud_init_raw_data,
               so global_params and app_params both will be the latest. So they are the value we desire too.
            '''
            hostvar = array.merge(cloud_init_raw_data, ansible_kv_config)
            hostvar = array.merge(hostvar, self._read_hosts_var())
            self._hostvar = hostvar
        return self._hostvar

    def get_app_params(self, key=None, fallback=None):
        app_params = self.get_host_var().get('app_params', {})
        if key is None:
            return app_params
        return app_params.get(key, fallback)

    def get_global_params(self, key=None, fallback=None):
        global_params = self.get_host_var().get('global_params', {})
        if key is None:
            return global_params
        return global_params.get(key, fallback)

    def reinstall_systemd_service(self, service_name, template_fname=None, payload=None):
        if template_fname is None:
            template_fname = service_name
        template_path = f'{self.app_templates_dir_for()}/{template_fname}.service'
        systemd.reinstall_systemd_service(service_name=service_name, template_path=template_path, payload=payload)
        self.logger.info('reinstall_systemd_service: %s => %s', template_path, service_name)

    @functocli.keep_method
    def check_health(self):
        print('%s is good' % (sys.argv[0]))



import sys
import shutil
import os

from os import path

from echobox.tool import file
from echobox.tool import strings


class RandomDir(object):

    def setUp(self):
        self.tests_root_dir = self.get_root_dir()
        name = strings.random_str(10)
        self.temp_dir = path.join(self.tests_root_dir, 'tmp/tmp-%s' % (name))
        file.ensure_dir(self.temp_dir)

    def get_root_dir(self):
        import __main__ as main
        self.is_interactive = hasattr(main, '__file__')
        if self.is_interactive:
            root_dir = path.dirname(path.realpath(sys.argv[0]))
            return root_dir
        else:
            root_dir = os.getcwd()
            return root_dir

    def random_str(self, count=10):
        return strings.random_str(count)

    def tearDown(self):
        shutil.rmtree(self.temp_dir)
        pass



import random
import unittest

from echobox.tool import cache
from echobox.tool import local_cache
from echobox.tool import redistool


class TestToolredis(unittest.TestCase):

    def setUp(self):
        url = 'redis://:%s@%s:%s' % ('vhash', '172.16.2.7', '6379')
        redistool.init_proxy(url)

    def test_fetch_list_from(self):
        test_result = {}

        times = 100
        ids = list(range(1, times))
        for id in ids:
            test_result[id] = random.random()

        def get_key_fn(id):
            return 'cache-test-from-' + str(id)

        def get_list_fn(ids):
            result_map = {}
            for id in ids:
                result_map[id] = test_result[id]
            return result_map

        keys = list(map(get_key_fn, ids))
        for key in keys:
            cache.delete(key)

        fetched_result = cache.fetch_list(ids, get_key_fn, get_list_fn, None, None, 600, True)
        for id, item in fetched_result.items():
            self.assertEqual(item[1], cache.FROM_SLOW_OP)

        fetched_result = cache.fetch_list(ids, get_key_fn, get_list_fn, None, None, 600, True)
        for id, item in fetched_result.items():
            if local_cache.cache_proxy.type == local_cache.CACHE_TYPE_EMPTY:
                self.assertEqual(item[1], cache.FROM_REDIS)
            else:
                self.assertEqual(item[1], cache.FROM_LOCAL_CACHE)

        fetched_result = cache.fetch_list(ids, get_key_fn, get_list_fn, None, None, 0, True)
        for id, item in fetched_result.items():
            self.assertEqual(item[1], cache.FROM_REDIS)
        for key in keys:
            cache.delete(key)

    def test_fetch_list_single(self):
        test_result = {}

        times = 100
        ids = list(range(1, times))
        for id in ids:
            test_result[id] = random.random()

        def get_key_fn(id):
            return 'cache-test-single-' + str(id)

        def get_list_fn(ids):
            result_map = {}
            for id in ids:
                result_map[id] = test_result[id]
            return result_map

        keys = list(map(get_key_fn, ids))

        for key in keys:
            cache.delete(key)

        fetched_result = cache.fetch_list(ids, get_key_fn, get_list_fn, None, None, 600)

        for _id in range(1, 1000):
            random_id = random.randint(1, times - 1)
            fetched_result = cache.fetch_list(random_id, get_key_fn, get_list_fn, None, None, 600)
            self.assertEqual(fetched_result, test_result[random_id])

        for key in keys:
            cache.delete(key)

    def test_fetch_list_batch(self):
        test_result = {}
        ids = list(range(1, 100))
        for id in ids:
            test_result[id] = random.random()

        def get_key_fn(id):
            return 'cache-test-batch' + str(id)

        def get_list_fn(ids):
            result_map = {}
            for id in ids:
                result_map[id] = test_result[id]
            return result_map

        keys = list(map(get_key_fn, ids))
        for key in keys:
            cache.delete(key)

        fetched_result = cache.fetch_list(ids, get_key_fn, get_list_fn, None, None, 600)
        self.assertEqual(fetched_result, test_result)

        for key in keys:
            cache.delete(key)

    def test_fetch(self):
        result = random.random()

        def get_fn(key):
            return result

        key = 'cache-test-fetch-' + str(random.random())
        cache.delete(key)

        fetched_result = cache.fetch(key, get_fn, None, None, 600)
        self.assertEqual(fetched_result, result)

        cache.delete(key)


if __name__ == '__main__':
    unittest.main()



import json


class StrEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, bytes):
            return str(obj, encoding='utf-8')
        return json.JSONEncoder.default(self, obj)



# coding: utf-8
import sys
import traceback

is_py2 = sys.version[0] == '2'
if is_py2:
    from Queue import Queue
else:
    from queue import Queue
from threading import Thread
from threading import Timer, Lock
from threading import Event

import copy
from echobox.tool import array
from echobox.tool import system
import logging
import time

logger = logging.getLogger()

debug_concurrent = False


def waiting(get_fn, check_should_wait_fn, on_waiting_fn, interval=5, wait_before_start=0):
    if wait_before_start:
        time.sleep(wait_before_start)
    while True:
        ret = get_fn()
        should_wait = check_should_wait_fn(ret)
        if should_wait:
            if callable(on_waiting_fn):
                on_waiting_fn(ret)
            time.sleep(interval)
        else:
            return ret


class SimpleThreadPool(object):
    def __init__(self, num_of_thread=2):
        self.queue = Queue()
        self.num_of_thread = num_of_thread
        self.ts = []

    def _do_work(self):
        if debug_concurrent:
            logger.debug('begin_do_work')
        while True and not self.queue.empty():
            try:
                fn, args, kwargs = self.queue.get()
                fn(*args, **kwargs)
            except Exception as e:
                exc_info = sys.exc_info()
                info = traceback.format_exception(*exc_info)
                del exc_info
                logger.error(''.join(info))
            finally:
                self.queue.task_done()

    def add_task(self, fn, *args, **kwargs):
        self.queue.put((fn, args, kwargs))

    def start(self):
        for i in range(self.num_of_thread):
            t = Thread(target=self._do_work)
            t.daemon = True
            t.start()
            self.ts.append(t)

    def wait(self):
        for t in self.ts:
            t.join()


class QThreadPool(object):

    def __init__(self, num_of_thread=2):
        self.queue = Queue()
        for i in range(num_of_thread):
            t = Thread(target=self._do_work)
            t.daemon = True
            t.start()

    def _do_work(self):
        if debug_concurrent:
            logger.debug('begin_do_work')
        while True:
            try:
                fn, args, kwargs = self.queue.get()
                fn(*args, **kwargs)
            except Exception as e:
                logger.error(e)
            finally:
                self.queue.task_done()

    def add_task(self, fn, *args, **kwargs):
        self.queue.put((fn, args, kwargs))

    def wait(self):
        self.queue.join()


class AsyncRepeatedTimer(object):

    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.start()

    def _run(self):
        self.function(*self.args, **self.kwargs)
        self.is_running = False
        self.start()

    def start(self):
        if not self.is_running:
            self.is_running = True
            self._timer = Timer(self.interval, self._run)
            self._timer.daemon = True
            self._timer.start()

    def stop(self):
        self._timer.cancel()
        self.is_running = False


class TRepeatedTimer(Thread):

    def __init__(self, func, interval, max_count=-1, py2_rough_time=True, args=(), kwargs={}):
        Thread.__init__(self)

        self.stop_flag = Event()
        self.interval = interval
        self.function = func
        self.args = args
        self.kwargs = kwargs
        self.count = 0
        self.py2_rough_time = py2_rough_time
        self.max_count = max_count

    def run(self):
        while not self.stop_flag.is_set():
            try:
                self.function(*self.args, **self.kwargs)
            except (KeyboardInterrupt, SystemExit):
                raise
            except Exception as e:
                exc_info = sys.exc_info()
                info = traceback.format_exception(*exc_info)
                del exc_info
                logger.error(''.join(info))
            if is_py2 and self.py2_rough_time:
                cmd = 'sleep %f' % (self.interval)
                system.shell_run(cmd)
            else:
                self.stop_flag.wait(self.interval)

            self.count += 1
            if (self.max_count > 0 and self.count >= self.max_count):
                self.stop()

    def stop(self):
        self.stop_flag.set()


class DelayBatchOp:

    def __init__(self, callback, timeout=3, batch_size=10):
        self.callback = callback
        self.timeout = timeout
        self.batch_size = batch_size
        self.data_list = []
        self.lock = Lock()
        self.skip_timer_check_for_one_round = False
        self.timer = AsyncRepeatedTimer(self.timeout, self._do_timer_check)
        self.pool = QThreadPool(4)

    def _do_timer_check(self):
        if debug_concurrent:
            logger.debug('_do_timer_check')
        if self.skip_timer_check_for_one_round:
            self.skip_timer_check_for_one_round = False
            if debug_concurrent:
                logger.debug('_do_timer_check skipped')
            return
        self.lock.acquire()
        self._do_check(True)
        self.lock.release()

    def _do_check(self, skip_batch_size_check_for_timer=False):
        if debug_concurrent:
            logger.debug('_do_check')

        size = len(self.data_list)
        reach_batch_size = size >= self.batch_size
        if reach_batch_size:
            self.skip_timer_check_for_one_round = True

        if reach_batch_size or (size > 0 and skip_batch_size_check_for_timer):
            list = copy.deepcopy(self.data_list)
            self.data_list = []
            for sub_list in array.chunks(list, self.batch_size):
                self.pool.add_task(self.callback, sub_list)

    def add_delay_item(self, item):
        if debug_concurrent:
            logger.debug('add_delay_item: %s', item)
        self.lock.acquire()
        self.data_list.append(item)
        self._do_check(False)
        self.lock.release()



import os
import sys
import platform

import re
import subprocess
from subprocess import CalledProcessError

from signal import signal, SIGPIPE, SIG_DFL

from echobox.tool import datatypes


def get_os_name():
    os_name = platform.system()
    map = {
        'Linux': 'linux',
        'Darwin': 'mac',
        'Windows': 'windows',
        'CYGWIN': 'cygwin',
        'MINGW': 'MinGw',
    }
    for name, sys_name in map.items():
        if os_name.startswith(name):
            return sys_name
    return 'unknown'


def available_cpu_count():
    """ Number of available virtual or physical CPUs on this system, i.e.
    user/real as output by time(1) when called with an optimally scaling
    userspace-only program"""

    # cpuset
    # cpuset may restrict the number of *available* processors
    try:
        m = re.search(r'(?m)^Cpus_allowed:\s*(.*)$',
                      open('/proc/self/status').read())
        if m:
            res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
            if res > 0:
                return res
    except IOError:
        pass

    # Python 2.6+
    try:
        import multiprocessing
        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass

    # https://github.com/giampaolo/psutil
    try:
        import psutil
        return psutil.cpu_count()  # psutil.NUM_CPUS on old versions
    except (ImportError, AttributeError):
        pass

    # POSIX


def memtotal_bytes():
    bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES')
    return bytes


def memtotal_mb():
    return memtotal_bytes() / (1024 ** 2)


# adapt to `commands.getstatusoutput`
def run_cmd(cmd, stdout=None):
    ret = 0
    try:
        output = subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
    except CalledProcessError as e:
        ret = e.returncode
        output = e.output
    output = output.decode('utf-8')
    return ret, output.rstrip('\n')


def run_cmd_ret(cmd):
    return run_cmd(cmd)[1]


def shell_run(cmd, keep_pipeline=True, exit_on_error=True, stdout=None, stderr=None):
    if datatypes.is_list_or_tuple(cmd):
        cmd = '\n' + ';\n'.join(cmd)
    ret = 0
    if keep_pipeline:
        # https://stackoverflow.com/questions/10479825/python-subprocess-call-broken-pipe
        ret = subprocess.call(cmd, shell=True, stdout=stdout, stderr=stderr,
                              preexec_fn=lambda: signal(SIGPIPE, SIG_DFL))
    else:
        ret = subprocess.call(cmd, shell=True, stdout=stdout, stderr=stderr)
    if ret != 0 and exit_on_error:
        sys.exit(ret)
    return ret



import socket
import struct
from echobox.tool import system


def get_ip_address(ifname):
    if system.get_os_name() == 'windows':
        import ifaddr
        adapters = ifaddr.get_adapters()
        for adapter in adapters:
            for ip in adapter.ips:
                if ip.nice_name == ifname and ip.is_IPv4:
                    return ip.ip
    else:
        try:
            import fcntl
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            return socket.inet_ntoa(fcntl.ioctl(
                s.fileno(),
                0x8915,  # SIOCGIFADDR
                struct.pack('256s', ifname.encode('utf-8')[:15])
            )[20:24])
        except IOError as e:
            raise e
        finally:
            s.close()
    return None


def get_hostname_ip():
    ip = socket.gethostbyname(socket.gethostname())
    return ip


def get_local_ip_address():
    try:
        return get_ip_address_udp()
    finally:
        return '127.0.0.1'


def get_ip_address_udp():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('8.8.8.8', 80))
    ret = s.getsockname()[0]
    s.close()
    return ret


def is_open(ip, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((ip, int(port)))
        return True
    except:
        pass
    finally:
        s.close()
    return False



import unittest
from echobox.tool import array


class TestArray(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_merge(self):
        a = {'first': {'all_rows': {'pass': 'dog', 'number': '1'}}}
        b = {'first': {'all_rows': {'fail': 'cat', 'number': '5'}}}
        self.assertEqual(array.merge(a, b), {'first': {'all_rows': {'pass': 'dog', 'fail': 'cat', 'number': '5'}}})

        a = {'first': {}}
        b = {'first': {'name': 'Maya'}}
        self.assertEqual(array.merge(a, b), b)

        a = {'first': {}}
        b = {'first': {'name': 'Maya'}}
        self.assertEqual(array.merge(b, a), b)


if __name__ == '__main__':
    unittest.main()



from echobox.tool import system
from echobox.tool import template

SYSTEMD_FOLDER = '/etc/systemd/system'


def reinstall_systemd_service(service_name, template_path, payload=None, exit_on_error=True):
    service_fpath = _build_service_fpath(service_name=service_name)

    template_payload = {'payload': payload}
    template.render_to_file(template_path, template_payload, service_fpath)

    system.shell_run(f'chmod 664 {service_fpath}')

    restart_systemd_service(service_name=service_name, exit_on_error=exit_on_error)


def restart_systemd_service(service_name, exit_on_error=True):
    cmd_list = [
        'systemctl daemon-reload',
        f'systemctl enable {service_name}',
        f'systemctl restart {service_name}',
        f'systemctl status {service_name} --no-pager'
    ]
    system.shell_run(cmd_list, exit_on_error=exit_on_error)


def remove_systemd_service(service_name):
    cmd_list = [
        f'systemctl disable {service_name}',
        f'systemctl stop {service_name}',
        f'systemctl status {service_name} --no-pager',
        f'unlink {_build_service_fpath(service_name)}',
    ]
    system.shell_run(cmd_list)


def _build_service_fpath(service_name):
    return f'{SYSTEMD_FOLDER}/{service_name}.service'



from datetime import datetime
from tzlocal import get_localzone

local_timezone = get_localzone()


def local_now():
    time = local_timezone.localize(datetime.now())
    return time


def local_now_ios8061_str():
    time = local_now().strftime('%Y-%m-%dT%H:%M:%S%z')
    return time



from echobox.tool import cache
from echobox.tool import utils


class EnvTool(object):

    def __init__(self, app):
        self.app = app
        fn = self.app.root_dir + '/.box-config.yml'
        self.box_config = utils.load_yaml(fn, {})

    def docker_url_by_env(self, image_name):
        group_names = self.app.get_group_names()
        if 'nexus-genesis' in group_names:
            return image_name
        if 'firelighter-server' in group_names:
            return self.docker_genesis_url(image_name)
        else:
            return self.docker_url(image_name)

    def docker_url(self, image_name=''):
        return self.box_config.get('docker_url_pre', '') + image_name

    def docker_publish_url(self, image_name=''):
        return self.box_config.get('docker_publish_url_pre', '') + image_name

    def docker_genesis_url(self, image_name=''):
        return self.box_config.get('docker_genesis_url_pre', '') + image_name

    def setup_redis_for_cache(self):
        config = self.box_config[self.app.env]
        redis = config['redis']
        redis_url = 'redis://:%s@%s:%s' % (redis['auth'], redis['host'], redis['port'])
        cache.set_redis_url(redis_url)



import oss2
import logging
import os
import time

from echobox.tool import file


class AliyunOss:

    def __init__(self, app, oss_config, dry_run=False):
        self.logger = app.logger

        aliyun_stupid_loggers = ['oss2', 'ask_queue.py', 'http.py', 'select_response.py', 'auth.py', 'utils.py',
                                 'xml_utils.py', 'resumable.py', 'api.py']
        for logger in aliyun_stupid_loggers:
            logging.getLogger(logger).setLevel(logging.CRITICAL)

        auth = oss2.Auth(oss_config['access_key_id'], oss_config['access_key_secret'])
        self.bucket = oss2.Bucket(auth, 'http://%s' % (oss_config['endpoint']), oss_config['bucket'])
        self.dry_run = dry_run

    def check_file_exist_with_md5(self, filepath, oss_path):
        etag = ''
        try:
            result = self.bucket.head_object(oss_path)
            etag = result.etag
        except Exception as e:
            pass
        md5_file = file.md5_file(filepath)
        if etag and md5_file and etag.lower() == md5_file.lower():
            return True
        else:
            return False

    def upload_if_not_exist(self, filepath, oss_path):
        if not os.path.isfile(filepath):
            self.logger.error('upload_if_not_exist: %s => %s, file not exist', filepath, oss_path)
            return
        if self.check_file_exist_with_md5(filepath, oss_path):
            self.logger.info('upload: %s => %s, already exist', filepath, oss_path)
            return
        self.logger.debug('upload_if_not_exist: %s => %s', filepath, oss_path)
        self.upload(filepath, oss_path)

    def upload(self, filepath, oss_path):
        if not self.dry_run:
            start = time.time()
            self.bucket.put_object_from_file(oss_path, filepath)
            self.logger.info('upload: %s => %s, took: %ss', filepath, oss_path, time.time() - start)
        else:
            self.logger.info('upload: %s => %s', filepath, oss_path)

    def download(self, oss_path, filepath):
        self.logger.info('download: %s => %s', oss_path, filepath)
        file.ensure_dir_for_file(filepath)
        if not self.dry_run:
            start = time.time()
            self.bucket.get_object_to_file(oss_path, filepath)
            self.logger.info('download: %s => %s, took: %ss', oss_path, filepath, time.time() - start)
        else:
            self.logger.info('download: %s => %s', oss_path, filepath)

    def download_if_not_exist(self, oss_path, filepath):
        if self.check_file_exist_with_md5(filepath, oss_path):
            self.logger.info('download: %s => %s, already exist', oss_path, filepath)
            return
        self.logger.debug('download_if_not_exist: %s => %s', oss_path, filepath)
        self.download(oss_path, filepath)

    def download_dir_if_not_exist(self, oss_prefix=None, local_dir='/tmp'):
        if oss_prefix and not oss_prefix.endswith('/'):
            oss_prefix = oss_prefix + '/'

        for obj in oss2.ObjectIterator(self.bucket, prefix=oss_prefix, delimiter='/'):
            name = obj.key[len(oss_prefix):] if oss_prefix else obj.key
            obj_local_path = os.path.join(local_dir, name)

            if obj.is_prefix():
                self.download_dir_if_not_exist(obj.key, obj_local_path)
            else:
                obj_local_path = os.path.join(local_dir, name)
                self.download_if_not_exist(obj.key, obj_local_path)



import unittest
from echobox.tool import dockerutil


class TestDockerUtil(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_all(self):
        ip = dockerutil.docker0_ip()


if __name__ == '__main__':
    unittest.main()



import unittest
from echobox.tool import strings


class TestUtils(unittest.TestCase):

    def test_all(self):
        map = {
            'a': 'A',
            '1.1': '1.1',
            'scaling_group': 'ScalingGroup',
        }
        for src, dst in map.items():
            self.assertEquals(strings.camelize_key(src), dst)


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime

from echobox.app import eventlog
from echobox.app.devops import DevOpsApp


class App(DevOpsApp):
    def __init__(self):
        DevOpsApp.__init__(self, 'cptool-test')


class TestToolUtils(unittest.TestCase):

    def setUp(self):
        self.app = App()

    def test_event_log(self):
        for i in range(1, 10):
            payload = {'time': str(datetime.now())}
            eventlog.add_event_log('test', payload)


if __name__ == '__main__':
    unittest.main()



import os
import json
import sys
import pprint
import jsonpickle
from ruamel import yaml
from jinja2 import Environment
from backports import configparser

from echobox.tool import file
from echobox.tool import array


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def load_properties(fn, fallback=None, to_dict=True):
    if not os.path.isfile(fn):
        return fallback
    data = None
    config_string = '[DEFAULT]\n' + file.file_get_contents(fn)
    cp = configparser.ConfigParser()
    cp.optionxform = str
    cp.read_string(config_string)
    return dict(cp['DEFAULT']) if to_dict else cp['DEFAULT']


def dump_properties(data, fn):
    cp = configparser.ConfigParser()
    cp.optionxform = str
    cp['DEFAULT'] = data
    with open(fn, 'w') as f:
        cp.write(f, space_around_delimiters=False)
    file_content = file.file_get_contents(fn)
    file_content = file_content.replace('[DEFAULT]\n', '').strip()
    file.file_put_contents(fn, file_content)


def load_json(fn, fallback=None):
    if not os.path.isfile(fn):
        return fallback
    data = None
    with open(fn, 'r') as f:
        data = json.load(f)
    return data if data is not None else fallback


def load_yaml(fn, fallback=None):
    if not os.path.isfile(fn):
        return fallback
    data = None
    with open(fn, 'r') as f:
        data = yaml.safe_load(f)
    return data if data is not None else fallback


def load_yaml_from_dir(dir, fallback=None):
    if not os.path.isdir(dir):
        return fallback
    file_list = file.listdir(dir)
    data = {}
    for path in file_list:
        if '.yml' not in path and '.yaml' not in path:
            continue
        data = array.merge(data, load_yaml(path, {}))
    return data


def load_yaml_config(config_file, config_section=None, extra_kvs=None):
    config = load_yaml(config_file, {})
    if config_section is not None:
        key_list = config_section.split('.')
        for sub_key in key_list:
            if sub_key not in config:
                raise Exception('Can not find config section for sub key:' + sub_key)
            config = config[sub_key]

    if extra_kvs is not None:
        config.update(extra_kvs)
    return config


def dump_json(data, fn, **kwargs):
    with open(fn, 'w') as outfile:
        json_str = jsonpickle.encode(data)  # for data contains object
        data = json.loads(json_str)
        json.dump(data, outfile, **kwargs)


def pjson(data):
    try:
        return json.dumps(data, indent=2, sort_keys=True)
    except:
        return pprint.pformat(object=data, indent=2, sort_dicts=True)  # for data contains object


def pyaml(data):
    yaml.Dumper.ignore_aliases = lambda *args: True
    return yaml.safe_dump(data, default_flow_style=False)


# TODO: should be dump_yaml(data, fn)
def dump_yaml(fn, data):
    with open(fn, 'w') as outfile:
        yaml.safe_dump(data, outfile, default_flow_style=False)


def json_to_yaml(src, out):
    data = load_json(src)
    dump_yaml(data, out)



import json
import logging
import time
from echobox.tool import timeutil

from echobox.app.appconfig import appconfig

event_logger = logging.getLogger('event-log')


def add_event_log(event_key, payload):
    payload['env'] = appconfig.get_env()
    msg = 'cp %s %s' % (event_key, json.dumps(payload))
    event_logger.info(msg)


def add_app_monitor_log(key, value, time=None):
    time = time if time else timeutil.local_now_ios8061_str()
    data = {
        'event_key': key,
        'event_value': value,
        'time': time,
    }
    payload = {}
    payload['env'] = appconfig.get_env()
    payload['log'] = json.dumps(data)
    msg = 'app-monitor monitor-log %s' % (json.dumps(payload))
    event_logger.info(msg)


def log_func_call(func, *args, **kwargs):
    def timed(*args, **kw):
        start = time.time() * 1000
        result = func(*args, **kw)
        payload = {}
        payload['name'] = func.__name__
        payload['rt'] = time.time() * 1000 - start
        add_event_log('func-call', payload)
        return result

    return timed



import unittest
from echobox.tool import net
from echobox.tool import dockerutil
from echobox.tool import system


class TestNet(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_all(self):
        if system.get_os_name() == 'mac':
            self.assertRaises(IOError, net.get_ip_address, 'docker0')
        else:
            ip1 = net.get_ip_address('docker0')
            ip2 = dockerutil.get_docker_network_gw('bridge')
            self.assertEqual(ip1, ip2)


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages

with open('requirements.txt', 'r') as f:
    lines = [line.strip() for line in f.readlines()]

install_requires = [line for line in lines if line and not line.startswith('#')]

name = 'echobox'
version = '1.5.38'
author = 'kiuber'
author_email = 'kiuber.zhang@gmail.com'

packages = find_packages()
print(f'packages: {packages}')

setup(
    name=name,
    version=version,
    keywords=[name, author],
    description=name,
    license='',
    install_requires=install_requires,

    scripts=[],

    author=author,
    author_email=author_email,
    url='',

    packages=packages,
    platforms='any',
    # https://pypi.org/classifiers/
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
    ]
)



import re
import six
from itertools import chain

try:  # Python 2
    items = 'iteritems'
except NameError:  # Python 3
    items = 'items'

_RaiseKeyError = object()


def is_list_or_tuple(value):
    if isinstance(value, six.string_types):
        return False
    return isinstance(value, (list, tuple))


def merge(src, update, list_behaviour='replace'):
    for key, value in update.items():
        if isinstance(value, dict):
            old = src.setdefault(key, {})
            merge(old, value, list_behaviour)
        elif isinstance(value, six.string_types):
            src[key] = value
        elif isinstance(value, tuple):
            old = src.setdefault(key, ())
            if list_behaviour == 'merge' and is_list_or_tuple(node):
                src[key] = old + value
            else:
                src[key] = value
        elif isinstance(value, list):
            old = src.setdefault(key, [])
            if list_behaviour == 'merge' and is_list_or_tuple(old):
                old.extend(value)
            else:
                src[key] = value
        else:
            src[key] = value
    return src


def filter_item(item, key, re_pattern):
    if callable(key):
        value = key(item)
        return _match_value(value, re_pattern)
    else:
        value = item[key]
        return _match_value(value, re_pattern)


def filter_item_by_dict(item, dict):
    for key, value in dict.items():
        if key not in item or item[key] != value:
            return False
    return True


def filter_by_keys(item, keys):
    ret = {}
    for key in keys:
        ret[key] = item[key]
    return ret


def _match_value(value, re_pattern):
    if isinstance(value, six.string_types):
        return re_pattern.match(value)
    if isinstance(value, (list, tuple)):
        for _item in value:
            if _match_value(_item, re_pattern):
                return True
    return False


class NameIdItem(dict):
    __slots__ = ('id_key', 'name_key')

    def __init__(self, id_key, name_key, *args, **kwargs):
        self.id_key = id_key
        self.name_key = name_key
        super(NameIdItem, self).__init__(*args, **kwargs)

    def get_id(self):
        return self[self.id_key]

    def get_name(self):
        return self[self.name_key]


class NameIdDict(dict):
    __slots__ = ('id_key', 'name_key', 'name_id_map')

    def __init__(self, id_key, name_key, *args, **kwargs):
        self.id_key = id_key
        self.name_key = name_key
        self.name_id_map = {}
        data_indexed_by_id = self._process_args(*args, **kwargs)
        super(NameIdDict, self).__init__(data_indexed_by_id)

    def _process_args(self, *args, **kwargs):
        data = dict(*args, **kwargs)
        data_indexed_by_id = {}
        for item in data.values():
            id = item[self.id_key]
            name = item[self.name_key]
            self.name_id_map[name] = id
            data_indexed_by_id[id] = NameIdItem(self.id_key, self.name_key, item)
        return data_indexed_by_id

    def _should_be_id(self, k):
        if k in self.name_id_map:
            return self.name_id_map[k]
        return k

    def __getitem__(self, k):
        return super(NameIdDict, self).__getitem__(self._should_be_id(k))

    def __setitem__(self, k, v):
        return super(NameIdDict, self).__setitem__(self._should_be_id(k), v)

    def __delitem__(self, k):
        k = self._should_be_id(k)
        return super(NameIdDict, self).__delitem__(k)

    def get(self, k, default=None):
        return super(NameIdDict, self).get(self._should_be_id(k), default)

    def setdefault(self, k, default=None):
        return super(NameIdDict, self).setdefault(self._should_be_id(k), default)

    def pop(self, k, v=_RaiseKeyError):
        if v is _RaiseKeyError:
            return super(NameIdDict, self).pop(self._should_be_id(k))
        return super(NameIdDict, self).pop(self._should_be_id(k), v)

    def update(self, *args, **kwargs):
        super(NameIdDict, self).update(self._process_args(*args, **kwargs))

    def __contains__(self, k):
        return super(NameIdDict, self).__contains__(self._should_be_id(k))

    def copy(self):  # don't delegate w/ super - dict.copy() -> dict :(
        return type(self)(self)

    @classmethod
    def fromkeys(cls, keys, v=None):
        return super(NameIdDict, cls).fromkeys((self._should_be_id(k) for k in keys), v)

    def __repr__(self):
        return '{0}({1})'.format(type(self).__name__, super(NameIdDict, self).__repr__())

    def get_id(self, id_or_name):
        if id_or_name in self:
            return self[id_or_name][self.id_key]

    def get_name(self, id_or_name):
        if id_or_name in self:
            return self[id_or_name][self.name_key]



def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


def merge(src, update):
    """
    >>> a = { 'first' : { 'all_rows' : { 'pass' : 'dog', 'number' : '1' } } }
    >>> b = { 'first' : { 'all_rows' : { 'fail' : 'cat', 'number' : '5' } } }
    >>> merge(a, b) == { 'first' : { 'all_rows' : { 'pass' : 'dog', 'fail' : 'cat', 'number' : '5' } } }
    True
    """
    for key, value in update.items():
        if isinstance(value, dict):
            # get node or create one
            node = src.setdefault(key, {})
            merge(node, value)
        else:
            src[key] = value

    return src


def fetch(data, keys):
    """
    >>> data = [ {'k1': '1v1', 'k2': '1v2', 'k3': '1v3'}, {'k1': '2v1', 'k2': '2v2', 'k3': '2v3'}]
    >>> keys = ['k1', 'k3']
    >>> fetch(data, keys)
    [{'k1': '1v1', 'k3': '1v3'}, {'k1': '2v1', 'k3': '2v3'}]
    """
    ret = []
    for item in data:
        ret.append({key: item[key] for key in keys})
    return ret


def filter_by_key_value(data, key, value, count=1):
    """
    >>> data = [ {'k1': '1v1', 'k2': '1v2', 'k3': '1v3'}, {'k1': '2v1', 'k2': '2v2', 'k3': '2v3'}]
    >>> key = 'k1'
    >>> keys = '2v1'
    >>> filter_by_key_value(data, key, value)
    {'k1': '2v1', 'k3': '2v3'}
    """
    ret = list(filter(lambda item: (item[key] == value), data))
    if count == 1 and len(ret) > 0:
        return ret[0]
    return ret



#!/usr/bin/env python

from echobox.app.env import EnvTool
from echobox.tool import functocli
from echobox.tool import template

opsbox_image = 'pipcli'


@functocli.keep_class
class PipCli(object):

    def __init__(self, app):
        self.app = app
        self.env_tool = EnvTool(app)
        self.opsbox_image = self.env_tool.docker_url(opsbox_image)
        self.opsbox_publish_image = self.env_tool.docker_publish_url(opsbox_image)

    def build_image(self):
        payload = {
            'opsbox_lib_version': self.app.get_global_params()['opsbox_lib_version'],
        }

        dst = self.app.app_runtime_storage_dir + '/opsbox-Dockerfile'
        self.app.template_to('opsbox-Dockerfile', dst, payload)

        cmd = 'cd %s; docker build -t %s -f %s .' % (self.app.app_runtime_storage_dir, self.opsbox_image, dst)
        self.app.shell_run(cmd)

    def push_image(self):
        self.shell_run('docker tag %s %s' % (self.opsbox_image, self.opsbox_publish_image))
        self.shell_run('docker push %s' % (self.opsbox_publish_image))

    def start_cli(self, args, cmd):
        cmd_data = {}
        cmd_data['image'] = self.opsbox_image
        cmd_data['args'] = args
        cmd_data['cmd'] = cmd
        cmd = template.render_str('docker run -d --restart always {{ args }} {{ image }} {{ cmd }}', cmd_data)
        self.app.shell_run(cmd)
